@@ -39,7 +39,8 @@ Boston, MA 02111-1307, USA.  */
 #include "eval.h"
 
 /* Node in a doubly-linked list of Perl references to Lisp objects.
-   Each such reference holds a pointer to a struct sv_lisp.  */
+   Each such reference holds a pointer to a struct sv_lisp.
+   They are linked together for marking during gc.  */
 struct sv_lisp
   {
     Lisp_Object obj;		/* must be first */
@@ -49,6 +50,7 @@ struct sv_lisp
 #define SV_LISP_NEXT(sv) (((struct sv_lisp *) SvIVX (sv))->next)
 #define SV_LISP_PREV(sv) (((struct sv_lisp *) SvIVX (sv))->prev)
 
+/* FIXME: I suspect we can do without struct perl_frame.  */
 struct perl_frame
   {
     /* Link to next outer stack frame.  */
@@ -78,11 +80,11 @@ struct perl_interpreter_object
   {
     PerlInterpreter *interp;
 
-    /* The int returned by perl_parse() or perl_run().  */
+    /* The int returned by perl_parse() or perl_run().  Not gcpro'd.  */
     Lisp_Object status;
 
     /* One of Qparsing, Qparsed, Qrunning, Qran, Qdestructing, Qdestructed,
-       Qbad.  */
+       and Qbad.  Not gcpro'd  */
     Lisp_Object phase;
 
     /* Doubly linked list of Lisp objects referenced in Perl (for GC).  */
@@ -101,6 +103,10 @@ struct perl_interpreter_object
 	struct perl_frame *p;
       } top;
 
+    /* The Emacs::Lisp::Object stash, cached for performance.
+       No ref is held, and god forbid anyone should delete it!  */
+    HV *elo_stash;
+
     /* Hack so that we know whether to return to perl_call_emacs_main
        when `kill-emacs' is called.  */
     char in_emacs_main;
@@ -109,7 +115,8 @@ struct perl_interpreter_object
 struct perl_interpreter_object *top_level_perl, *current_perl;
 EXFUN (Fperl_destruct, 1);
 static Lisp_Object internal_perl_call P_ ((int nargs, Lisp_Object *args,
-					   int do_eval));
+					   int do_eval,
+					   Lisp_Object (*conv) P_ ((SV *sv))));
 static void init_perl ();
 
 #define INIT_PERL do { if (!current_perl) init_perl (); } while (0)
@@ -127,17 +134,33 @@ static void init_perl ();
    ((struct perl_interpreter_object *) (XFOREIGN_OBJECT (obj)->data))
 
 /* How to tell when a Perl object's interpreter is no longer usable.  */
-#define LISP_XV_INTERP(obj)					\
+#define LISP_SV_INTERP(obj)					\
    (XPERL_INTERPRETER (XFOREIGN_OBJECT (obj)->lisp_data))
 #define PERL_DEFUNCT_P(perl)			\
    (EQ ((perl)->phase, Qdestructed)		\
     || EQ ((perl)->phase, Qbad))
 
-static Lisp_Object perl_interpreter;
+#define CACHE_ELO_STASH(perl)						\
+   { if (! perl->elo_stash)						\
+       perl->elo_stash = Perl_gv_stashpv ("Emacs::Lisp::Object", 1); }
+#define BARE_SV_LISPP(sv)						 \
+   (SvOBJECT (sv) && SvSTASH (sv) == PERLMACS_SV_INTERP (sv)->elo_stash)
+#define XBARE_SV_LISP(sv) (*((Lisp_Object *) SvIVX (sv)))
+#define FAST_SV_LISPP(sv) (SvROK (sv) && BARE_SV_LISPP (SvRV (sv)))
+
+#define LISP_SVP(obj)							  \
+   (FOREIGN_OBJECTP (obj) && XFOREIGN_OBJECT (obj)->vptr == &generic_vtbl)
+#define XLISP_SV(obj)				\
+   ((SV *) XFOREIGN_OBJECT (obj)->data)
+
+#define PERL_OBJECT_P(obj)					\
+   (FOREIGN_OBJECTP (obj)					\
+    && XFOREIGN_OBJECT (obj)->vptr->type_of == lisp_sv_type_of)
+
+
+static Lisp_Object perl_interpreter, Qperl_interpreter;
 static Lisp_Object Qmake_perl_interpreter, Qperl_error;
-static Lisp_Object Qperl_interpreter, Qperl_object, Qperl_scalar;
-static Lisp_Object Qperl_array, Qperl_hash, Qperl_code, Qperl_glob;
-static Lisp_Object Qperl_call, Qvoid_context, Qscalar_context, Qlist_context;
+static Lisp_Object Qvoid_context, Qscalar_context, Qlist_context;
 Lisp_Object Qparsing, Qparsed, Qran, Qdestructed, Qbad;
 
 static struct Lisp_Foreign_Object_VTable interp_vtbl;
@@ -162,9 +185,7 @@ pop_lisp_frame (interp)
   return make_number (0);
 }
 
-/* Frame is Perl.
-   (Lisp.xs:funcall or Emacs::Lisp::Object::funcall,
-   or perlmacs_call_emacs_main).
+/* Frame is Perl (perlmacs_funcall or perlmacs_call_emacs_main).
    Enter a new Lisp frame.  */
 static void
 push_lisp_frame (lfp)
@@ -200,7 +221,7 @@ perlmacs_call_emacs_main (argc, argv, envp)
   return XINT (perlmacs_catch.val);
 }
 
-/* Frame is start, either Lisp or Perl (emacs.c:Fkill_emacs).  */
+/* Frame is none or Perl (emacs.c:Fkill_emacs).  */
 void
 perlmacs_exit (status)
      int status;
@@ -240,6 +261,7 @@ perlmacs_init ()
 
       init_lisp (1, dummy_argv, 0);
     }
+  CACHE_ELO_STASH (current_perl);
 }
 
 /* Frame is none (emacs.c:emacs_main)
@@ -304,7 +326,8 @@ perlmacs_error_handler (e)
   return make_number (0);
 }
 
-/* Frame is Perl.  */
+/* Frame is Perl
+   (Lisp.xs: Emacs::Lisp::funcall or Emacs::Lisp::Object::funcall).  */
 Lisp_Object
 perlmacs_funcall (nargs, args)
      int nargs;
@@ -426,8 +449,7 @@ push_perl_frame (struct perl_frame *pfp, char need_jmpenv)
   /* Refuse to enter Perl if Emacs is shutting down, since data structures
      are likely to be in an inconsistent state.  */
   if (EQ (emacs_phase, Qdestructing))
-    error ("Attempt to enter Perl during global destruction.  Call"
-	   " `perl-destruct' before exiting Emacs.");
+    error ("Attempt to enter Perl while exiting Emacs");
 
   pfp->prev = current_perl->top.l;
   current_perl->top.p = pfp;
@@ -531,9 +553,10 @@ abort_perl_call (jumpret)
    Convert the value or values (if any) returned by `perl-call' or `perl-eval'
    to a Lisp object.  */
 static Lisp_Object
-perlmacs_retval (ctx, numret)
+perlmacs_retval (ctx, numret, conv)
      I32 ctx;
      I32 numret;
+     Lisp_Object (*conv) P_ ((SV *sv));
 {
   Lisp_Object *vals;
   int i;
@@ -545,7 +568,7 @@ perlmacs_retval (ctx, numret)
     case G_SCALAR:
       {
 	dTHR;
-	return perlmacs_sv_to_lisp (*PL_stack_sp);
+	return (*conv) (*PL_stack_sp);
       }
     default:  /* G_ARRAY */
       if (numret == 0)
@@ -555,7 +578,7 @@ perlmacs_retval (ctx, numret)
 	PL_stack_sp -= numret;
 	vals = (Lisp_Object *) alloca (numret * sizeof (Lisp_Object));
 	for (i = 0; i < numret; i++)
-	  vals [i] = perlmacs_sv_to_lisp (PL_stack_sp [i+1]);
+	  vals [i] = (*conv) (PL_stack_sp [i+1]);
 	return Flist (numret, vals);
       }
     }
@@ -566,17 +589,21 @@ enum perl_entry
     PERL_CALL_SV,
     PERL_EVAL_SV,
     SV_FREE,
-    EVAL_AND_CALL
+    EVAL_AND_CALL,
+    CALL_RAW
   };
 
 /* Frame is Perl (perlmacs.c, multiple functions).
    Call the specified Perl API function and pop to the previous Lisp frame.  */
+/* Note: Perl 5.005 requires stdarg.  */
 static Lisp_Object
-call_perl (entry, sv, ctx)
-     enum perl_entry entry;
-     SV *sv;
-     I32 ctx;
+call_perl (enum perl_entry entry, ...)
 {
+  va_list ap;
+  SV *sv;
+  I32 ctx;
+  Lisp_Object (*conv) P_ ((SV *sv));
+  char *str;
   dJMPENV;
   dTHR;
   int ret;
@@ -598,25 +625,37 @@ call_perl (entry, sv, ctx)
   if (ret)
     abort_perl_call (ret);
 
+  va_start (ap, entry);
   switch (entry)
     {
     case PERL_EVAL_SV:
+      sv = va_arg (ap, SV *);
+      ctx = va_arg (ap, I32);
+      conv = va_arg (ap, void *);
+
       numret = perl_eval_sv (sv, ctx);
       if (SvTRUE (ERRSV))
 	abort_perl_call (3);
-      retval = perlmacs_retval (ctx, numret);
+      retval = perlmacs_retval (ctx, numret, conv);
       break;
 
     case PERL_CALL_SV:
+      sv = va_arg (ap, SV *);
+      ctx = va_arg (ap, I32);
+      conv = va_arg (ap, void *);
+
       numret = perl_call_sv (sv, ctx);
-      retval = perlmacs_retval (ctx, numret);
+      retval = perlmacs_retval (ctx, numret, conv);
       break;
 
     case EVAL_AND_CALL:
-      sv = perl_eval_pv ((char *) sv, 1);
+      str = va_arg (ap, char *);
+      ctx = va_arg (ap, I32);
+
+      sv = perl_eval_pv (str, 1);
       POPMARK;
       if (! SvOK (sv))
-	/* XXX The second argument to perl_eval_pv ("croak_on_error")
+	/* FIXME: The second argument to perl_eval_pv ("croak_on_error")
 	   doesn't seem to work for syntax errors.  I don't know how
 	   to distinguish syntax errors from expressions that evaluate
 	   to undef.  */
@@ -624,14 +663,17 @@ call_perl (entry, sv, ctx)
 		    " possible syntax error");
 
       numret = perl_call_sv (sv, ctx);
-      retval = perlmacs_retval (ctx, numret);
+      retval = perlmacs_retval (ctx, numret, perlmacs_sv_to_lisp);
       break;
 
     case SV_FREE:
+      sv = va_arg (ap, SV *);
+
       Perl_sv_free (sv);
       retval = Qnil;
       break;
     }
+  va_end (ap);
 
   /* Indicate to perlmacs_funcall that there were no errors.  */
   lfp = current_perl->top.p->prev;
@@ -847,6 +889,7 @@ new_perl (in_lisp)
   else
     perl->top.p = &perl->start.p;
 
+  perl->status = make_number (0);
   return perl;
 }
 
@@ -913,7 +956,10 @@ Create and return a new Perl interpreter object.")
 	}
     }
   else
-    argv = def_argv;
+    {
+      argv = def_argv;
+      nargs = sizeof def_argv / sizeof def_argv [0];
+    }
 
   /* Test whether the user pressed ^G before entering a potentially
      long-lasting bit of Perl.  */
@@ -946,6 +992,7 @@ Create and return a new Perl interpreter object.")
     perl->phase = Qparsed;
   UNGCPRO;
 
+  CACHE_ELO_STASH (perl);
   return unbind_to (count, this_perl);
 }
 
@@ -1044,6 +1091,31 @@ If no arg is given, shut down the current Perl interpreter.")
   return Qnil;
 }
 
+DEFUN ("perl-status", Fperl_status, Sperl_status, 0, 1, 0,
+  "Return what would be perl's exit status.")
+  (interpreter)
+     Lisp_Object interpreter;
+{
+  if (NILP (interpreter))
+    interpreter = perl_interpreter;
+
+  CHECK_PERL (interpreter);
+  return XPERL_INTERPRETER (interpreter)->status;
+}
+
+DEFUN ("perl-phase", Fperl_phase, Sperl_phase, 0, 1, 0,
+  "Return the Perl interpreter's phase, one of the following symbols:\n\
+parsing, parsed, running, ran, destructing, destructed, bad.")
+  (interpreter)
+     Lisp_Object interpreter;
+{
+  if (NILP (interpreter))
+    interpreter = perl_interpreter;
+
+  CHECK_PERL (interpreter);
+  return XPERL_INTERPRETER (interpreter)->phase;
+}
+
 /* Fire up a Perl interpreter for use with Lisp functions that act
    implicitly on one.
    Don't call this function directly.  Use INIT_PERL.  */
@@ -1060,45 +1132,45 @@ init_perl ()
    methods, so long as they are called only internally.  */
 
 static Lisp_Object
-lisp_xv_type_of (object)
+lisp_sv_type_of (object)
      Lisp_Object object;
 {
+  char *str;
   SV *sv = (SV *) XFOREIGN_OBJECT (object)->data;
 
   /* Look to sv.c:sv_reftype for inspiration.  */
 
   if (SvOBJECT(sv))
-    /* FIXME: Let's return something more interesting.  */
-    return Qperl_object;
+    /* FIXME: Let's return something more interesting,
+       like the object's stash.  */
+    return intern ("perl-object");
 
-  /* In the places where I call `intern' I've never seen that value, and
-     I don't know whether it can happen under normal conditions.
-     (`perl-unknown' tends to indicate a bug.)  */
   switch (SvTYPE (sv))
     {
-      /* FIXME: Is `SVt_NULL' undef?
-	 We should perhaps return `perl-undef' if it is.  */
     case SVt_NULL:
     case SVt_IV:
     case SVt_NV:
     case SVt_RV:
-    case SVt_PV:	return Qperl_scalar;
+    case SVt_PV:
     case SVt_PVIV:
     case SVt_PVNV:
     case SVt_PVMG:
-    case SVt_PVBM:	return SvROK (sv) ? intern ("perl-ref") : Qperl_scalar;
-    case SVt_PVLV:	return intern ("perl-lvalue");
-    case SVt_PVAV:	return Qperl_array;
-    case SVt_PVHV:	return Qperl_hash;
-    case SVt_PVCV:	return Qperl_code;
-    case SVt_PVGV:	return Qperl_glob;
-    case SVt_PVFM:	return intern ("perl-format");
-    default:		return intern ("perl-unknown");
+    case SVt_PVBM:	str = SvROK (sv) ? "perl-ref" : "perl-scalar"; break;
+    case SVt_PVLV:	str = "perl-lvalue";	break;
+    case SVt_PVAV:	str = "perl-array";	break;
+    case SVt_PVHV:	str = "perl-hash";	break;
+    case SVt_PVCV:	str = "perl-code";	break;
+    case SVt_PVGV:	str = "perl-glob";	break;
+    case SVt_PVFM:	str = "perl-format";	break;
+    case SVt_PVIO:	str = "perl-io";	break;
+    default:		str = "perl-unknown";	break;
     }
+
+  return intern (str);
 }
 
 static void
-lisp_xv_mark (object)
+lisp_sv_mark (object)
      Lisp_Object object;
 {
   /* Mark the interpreter.  */
@@ -1106,7 +1178,7 @@ lisp_xv_mark (object)
 }
 
 static void
-lisp_xv_destroy (object)
+lisp_sv_destroy (object)
      Lisp_Object object;
 {
   struct perl_interpreter_object *perl;
@@ -1120,7 +1192,7 @@ lisp_xv_destroy (object)
   if (EQ (emacs_phase, Qdestructing))
     return;
 
-  perl = LISP_XV_INTERP (object);
+  perl = LISP_SV_INTERP (object);
   if (PERL_DEFUNCT_P (perl)
       || EQ (perl->phase, Qdestructing))
     return;
@@ -1132,14 +1204,14 @@ lisp_xv_destroy (object)
       SvREFCNT_dec (sv);
       return;
     }
-  call_perl (SV_FREE, sv, G_VOID);
+  call_perl (SV_FREE, sv);
 }
 
 static struct Lisp_Foreign_Object_VTable generic_vtbl =
   {
-    lisp_xv_type_of,	/* TYPE-OF method.  */
-    lisp_xv_destroy,	/* Destructor.  */
-    lisp_xv_mark,	/* MARK method.  Mark the interpreter.  */
+    lisp_sv_type_of,	/* TYPE-OF method.  */
+    lisp_sv_destroy,	/* Destructor.  */
+    lisp_sv_mark,	/* MARK method.  Mark the interpreter.  */
     0,			/* TO-STRING method.  FIXME: Write one!  */
     0,			/* EQUAL method.  FIXME: Write one!  */
     0			/* CALL method.  Invalid as a function.  */
@@ -1154,8 +1226,8 @@ lisp_cv_call (self, nargs, args)
   Lisp_Object *new_args;
   int new_nargs = nargs + 2;
 
-  if (PERL_DEFUNCT_P (LISP_XV_INTERP (self)))
-    error ("Attempt to call destroyed Perl code");
+  if (PERL_DEFUNCT_P (LISP_SV_INTERP (self)))
+    error ("Attempt to call Perl code in a destructed interpreter");
 
   new_args = (Lisp_Object *) alloca ((nargs + 2) * sizeof (Lisp_Object));
   new_args [0] = self;
@@ -1164,21 +1236,21 @@ lisp_cv_call (self, nargs, args)
       || EQ (args [0], Qvoid_context))
     {
       bcopy (args, new_args + 1, nargs * sizeof args[0]);
-      return internal_perl_call (nargs + 1, new_args, 0);
+      return internal_perl_call (nargs + 1, new_args, 0, perlmacs_sv_to_lisp);
     }
   else
     {
       new_args [1] = Qscalar_context;
       bcopy (args, new_args + 2, nargs * sizeof args[0]);
-      return internal_perl_call (nargs + 2, new_args, 0);
+      return internal_perl_call (nargs + 2, new_args, 0, perlmacs_sv_to_lisp);
     }
 }
 
 static struct Lisp_Foreign_Object_VTable code_vtbl =
   {
-    lisp_xv_type_of,	/* TYPE-OF method.  */
-    lisp_xv_destroy,	/* Destructor.  Use SV's.  */
-    lisp_xv_mark,	/* MARK method.  Mark the interpreter.  */
+    lisp_sv_type_of,	/* TYPE-OF method.  */
+    lisp_sv_destroy,	/* Destructor.  Use SV's.  */
+    lisp_sv_mark,	/* MARK method.  Mark the interpreter.  */
     0,			/* TO-STRING method.  Use default.  */
     0,			/* EQUAL method.  Use default.  */
     lisp_cv_call	/* CALL method.  */
@@ -1188,6 +1260,9 @@ Lisp_Object
 perlmacs_lisp_wrap_sv (sv)
      SV *sv;
 {
+  if (SvROK (sv))
+    sv = SvRV (sv);
+
   SvREFCNT_inc (sv);
   return new_foreign_object (SvTYPE (sv) == SVt_PVCV
 			     ? &code_vtbl : &generic_vtbl,
@@ -1198,11 +1273,8 @@ Lisp_Object
 perlmacs_sv_to_lisp (sv)
      SV *sv;
 {
-  STRLEN len;
-  char *pc;
-  /* We should gcpro ret if perlmacs_sv_to_lisp() becomes able to
-     call user code.  */
   Lisp_Object retval;
+  SV *sv1, *sv2;
 
   if (! sv)
     return Qnil;  /* grrr.... */
@@ -1217,85 +1289,97 @@ perlmacs_sv_to_lisp (sv)
     return make_float (SvNV (sv));
 #endif
 
-  if (SV_LISPP (sv))
-    return XSV_LISP (sv);
-
-  if (SvROK (sv))
+  if (SvPOK (sv))
     {
-      svtype type;
+      STRLEN len;
+      char *pc = SvPV (sv, len);
+      return make_unibyte_string (pc, (int) len);
+    }
 
-      sv = SvRV (sv);
-      type = SvTYPE (sv);
+  /* Don't distinguish, e.g., AV* from RV->AV*.  Perl can't pass us an AV*,
+     but this way, C can avoid allocating RV's.  */
+  sv1 = SvROK (sv) ? SvRV (sv) : sv;
 
+  if (BARE_SV_LISPP (sv1))
+    return XBARE_SV_LISP (sv1);
+
+  switch (SvTYPE (sv1))
+    {
+    case SVt_NULL:
+    case SVt_IV:
+    case SVt_NV:
+    case SVt_RV:
+    case SVt_PV:
+    case SVt_PVIV:
+    case SVt_PVNV:
+    case SVt_PVMG:
+    case SVt_PVBM:
+    case SVt_PVLV:
+    case SVt_PVHV:
+    case SVt_PVCV:
+    case SVt_PVFM:
+    case SVt_PVIO:
+      if (SvROK (sv1))
+	{
+	  sv2 = SvRV (sv1);
+
+	  /* Convert array ref ref to vector.  */
+	  if (SvTYPE (sv2) == SVt_PVAV)
+	    {
+	      SV **ary;
+	      I32 key;
+	      Lisp_Object *ptr;
+
+	      ary = AvARRAY ((AV *) sv2);
+	      key = AvFILLp ((AV *) sv2) + 1;
+	      retval = Fmake_vector (make_number (key), Qnil);
+	      ptr = XVECTOR (retval)->contents;
+	      while (key--)
+		ptr [key] = perlmacs_sv_to_lisp (ary [key]);
+
+	      return retval;
+	    }
+	}
+      break;
+
+    case SVt_PVAV:
+      /* Convert arrayref to list.  */
+      {
+	SV **ary;
+	I32 key;
+
+	ary = AvARRAY((AV *) sv1);
+	key = AvFILLp((AV *) sv1) + 1;
+	retval = Qnil;
+	while (key)
+	  retval = Fcons (perlmacs_sv_to_lisp (ary [--key]), retval);
+
+	return retval;
+      }
+      break;
+
+    case SVt_PVGV:
       /* Check for glob in package main.  */
-      if (type == SVt_PVGV
-	  /* FIXME: Perl often checks its own pointers for NULL-ness,
-	     but it may not be necessary here.  */
-	  && GvSTASH (sv)
-	  && HvNAME (GvSTASH (sv))
-	  && strEQ (HvNAME (GvSTASH (sv)), "main"))
+      if (GvSTASH (sv1) == PL_defstash)
 	{
 	  /* Convert globref to symbol.  */
+	  char *pc;
 	  const char *pcc;
 	  const char *end;
 	  char *name;
 
-	convert_glob:
-	  end = GvNAME (sv) + GvNAMELEN (sv);
-	  name = alloca (1 + GvNAMELEN (sv));
-	  for (pcc = GvNAME (sv), pc = name; pcc != end; ++pcc, ++pc)
+	  end = GvNAME (sv1) + GvNAMELEN (sv1);
+	  name = alloca (1 + GvNAMELEN (sv1));
+	  for (pcc = GvNAME (sv1), pc = name; pcc != end; ++pcc, ++pc)
 	    *pc = (*pcc == '_') ? '-' : (*pcc == '-') ? '_' : *pcc;
 	  *pc = '\0';
 
 	  return intern (name);
 	}
-
-      /* Convert arrayref to list.  */
-      if (type == SVt_PVAV)
-	{
-	  SV **ary;
-	  I32 key;
-
-	  ary = AvARRAY((AV *) sv);
-	  key = AvFILLp((AV *) sv) + 1;
-	  retval = Qnil;
-	  while (key)
-	    retval = Fcons (perlmacs_sv_to_lisp (ary [--key]), retval);
-
-	  return retval;
-	}
-
-      /* Convert array ref ref to vector.  */
-      if (SvROK (sv)
-	  && (type = SvTYPE (sv = SvRV (sv))) == SVt_PVAV)
-	{
-	  SV **ary;
-	  I32 key;
-	  Lisp_Object *ptr;
-
-	  ary = AvARRAY((AV *) sv);
-	  key = AvFILLp((AV *) sv) + 1;
-	  retval = Fmake_vector (make_number (key), Qnil);
-	  ptr = XVECTOR (retval)->contents;
-	  while (key--)
-	    ptr [key] = perlmacs_sv_to_lisp (ary [key]);
-
-	  return retval;
-	}
-
-      return perlmacs_lisp_wrap_sv (sv);
+      break;
     }
-#if 0
-  /* &func(*sym) doesn't seem to work unless sym appears elsewhere in Perl.
-     Globs are a mystery to me.  Better to disable this kind of stuff.  */
-  if (SvTYPE (sv) == SVt_PVGV && GvSTASH (sv)
-      && HvNAME (GvSTASH (sv))
-      && strEQ (HvNAME (GvSTASH (sv)), "main"))
-    goto convert_glob;
-#endif
 
-  pc = SvPV (sv, len);
-  return make_unibyte_string (pc, (int) len);
+  return perlmacs_lisp_wrap_sv (sv);
 }
 
 SV *
@@ -1316,8 +1400,8 @@ perlmacs_sv_wrap_lisp (obj)
     SV_LISP_PREV (current_perl->protlist) = sv;
   current_perl->protlist = sv;
 
-  rv = Perl_sv_bless (Perl_newRV_noinc (sv),
-		      Perl_gv_stashpv ("Emacs::Lisp::Object", 1));
+  /* FIXME: I doubt whether I'm adequately freeing these rv's.  */
+  rv = Perl_sv_bless (Perl_newRV_noinc (sv), current_perl->elo_stash);
   return rv;
 }
 
@@ -1327,99 +1411,131 @@ perlmacs_lisp_to_sv (obj)
 {
   if (NILP (obj))
     return &PL_sv_undef;
-  else
-    switch (XTYPE (obj))
-      {
-      case Lisp_Int:
-	return Perl_newSViv ((IV) XINT (obj));
+
+  /* Unconvert wrapped Perl data.  */
+  if (PERL_OBJECT_P (obj))
+    return Perl_newRV (XLISP_SV (obj));
+
+  switch (XTYPE (obj))
+    {
+    case Lisp_Int:
+      return Perl_newSViv ((IV) XINT (obj));
 
 #ifdef LISP_FLOAT_TYPE
-      case Lisp_Float:
-	return Perl_newSVnv (XFLOAT (obj)->data);
+    case Lisp_Float:
+      return Perl_newSVnv (XFLOAT (obj)->data);
 #endif
-      case Lisp_String:
-	return Perl_newSVpvn (XSTRING (obj)->data,
-			      (STRLEN) XSTRING (obj)->size);
+    case Lisp_String:
+      return Perl_newSVpvn (XSTRING (obj)->data,
+			    (STRLEN) XSTRING (obj)->size);
 
-      case Lisp_Symbol:
-	{
-	  /* Convert symbol to globref.  */
-	  EMACS_INT len = XSYMBOL (obj)->name->size;
-	  char *name = (char *) alloca (len + 3);
-	  char *pc = name;
-	  const char *pcc = &(XSYMBOL (obj)->name->data [0]);
-	  const char *end = pcc + len;
+    case Lisp_Symbol:
+      {
+	/* Convert symbol to globref.  */
+	EMACS_INT len = XSYMBOL (obj)->name->size;
+	char *name = (char *) alloca (len + 3);
+	char *pc = name;
+	const char *pcc = &(XSYMBOL (obj)->name->data [0]);
+	const char *end = pcc + len;
 
-	  *pc++ = ':';
-	  *pc++ = ':';
-	  for (; pcc != end; ++pcc, ++pc)
-	    *pc = (*pcc == '_') ? '-' : (*pcc == '-') ? '_' : *pcc;
-	  *pc = '\0';
-	  return Perl_newRV ((SV *) Perl_gv_fetchpv (name, 1, SVt_PVGV));
-	}
-      case Lisp_Cons:
-	{
-	  I32 i;
-	  Lisp_Object list;
-	  for (i = 1, list = XCDR (obj); 1; i++, list = XCDR (list))
-	    {
-	      if (NILP (list))
-		{
-		  /* Convert list to arrayref.  */
-		  AV *av = Perl_newAV ();
-
-		  Perl_av_extend (av, i);
-		  for (i = 0, list = obj; !NILP (list);
-		       i++, list = XCDR (list))
-		    Perl_av_store (av, i, perlmacs_lisp_to_sv (XCAR (list)));
-		  return Perl_newRV_noinc ((SV *) av);
-		}
-	      if (!CONSP (list))
-		break;
-	    }
-	}
-	break;
-      case Lisp_Vectorlike:
-	if (VECTORP (obj))
-	  {
-	    /* Convert vector to array ref ref.  */
-	    int i = XVECTOR (obj)->size;
-	    Lisp_Object *ptr = XVECTOR (obj)->contents;
-	    AV *av = Perl_newAV ();
-
-	    Perl_av_extend (av, i);
-	    while (i--)
-	      Perl_av_store (av, i, perlmacs_lisp_to_sv (ptr [i]));
-	    return Perl_newRV_noinc (Perl_newRV_noinc ((SV *) av));
-	  }
-	break;
-      case Lisp_Misc:
-	switch (XMISCTYPE (obj))
-	  {
-	  case Lisp_Misc_Foreign_Object:
-	    {
-	      /* Unconvert wrapped Perl data.  */
-	      struct Lisp_Foreign_Object *perl = XFOREIGN_OBJECT (obj);
-	      if (perl->vptr->type_of == lisp_xv_type_of)
-		return Perl_newRV ((SV *) perl->data);
-	    }
-	    break;
-	  }
-	break;
-
-	/* FIXME: What other object types are intelligently wrappable?  */
-
-      default:
-	break;
+	*pc++ = ':';
+	*pc++ = ':';
+	for (; pcc != end; ++pcc, ++pc)
+	  *pc = (*pcc == '_') ? '-' : (*pcc == '-') ? '_' : *pcc;
+	*pc = '\0';
+	return Perl_newRV ((SV *) Perl_gv_fetchpv (name, 1, SVt_PVGV));
       }
+    case Lisp_Cons:
+      {
+	I32 i;
+	Lisp_Object list;
+	for (i = 1, list = XCDR (obj); 1; i++, list = XCDR (list))
+	  {
+	    if (NILP (list))
+	      {
+		/* Convert list to arrayref.  */
+		AV *av = Perl_newAV ();
+
+		Perl_av_extend (av, i);
+		for (i = 0, list = obj; !NILP (list);
+		     i++, list = XCDR (list))
+		  Perl_av_store (av, i, perlmacs_lisp_to_sv (XCAR (list)));
+		return Perl_newRV_noinc ((SV *) av);
+	      }
+	    if (!CONSP (list))
+	      break;
+	  }
+      }
+      break;
+    case Lisp_Vectorlike:
+      if (VECTORP (obj))
+	{
+	  /* Convert vector to array ref ref.  */
+	  int i = XVECTOR (obj)->size;
+	  Lisp_Object *ptr = XVECTOR (obj)->contents;
+	  AV *av = Perl_newAV ();
+
+	  Perl_av_extend (av, i);
+	  while (i--)
+	    Perl_av_store (av, i, perlmacs_lisp_to_sv (ptr [i]));
+	  return Perl_newRV_noinc (Perl_newRV_noinc ((SV *) av));
+	}
+      break;
+
+      /* FIXME: What other object types are intelligently wrappable?  */
+
+    default:
+      break;
+    }
   return perlmacs_sv_wrap_lisp (obj);
 }
 
+DEFUN ("perl-object-p", Fperl_object_p, Sperl_object_p, 1, 1, 0,
+  "Return t if OBJECT is any type of Perl data or code.")
+     (object)
+     Lisp_Object object;
+{
+  if (PERL_OBJECT_P (object))
+    return Qt;
+  return Qnil;
+}
+
+DEFUN ("perl-to-lisp", Fperl_to_lisp, Sperl_to_lisp, 1, 1, 0,
+  "Return OBJECT as the corresponding Lisp type.\n\
+If the object is not Perl data or cannot be converted to Lisp,\n\
+it is returned unchanged.")
+     (object)
+     Lisp_Object object;
+{
+  if (LISP_SVP (object))
+    return perlmacs_sv_to_lisp (XLISP_SV (object));
+  return object;
+}
+
+DEFUN ("perl-wrap", Fperl_wrap, Sperl_wrap, 1, 1, 0,
+  "Return OBJECT as a blessed reference of Perl package Emacs::Lisp::Object.")
+     (object)
+     Lisp_Object object;
+{
+  return perlmacs_lisp_wrap_sv (perlmacs_sv_wrap_lisp (object));
+}
+
+DEFUN ("make-perl-data", Fmake_perl_data, Smake_perl_data, 1, 1, 0,
+  "Return OBJECT as the corresponding Perl type.\n\
+If the object is not Perl data or cannot be converted to Perl,\n\
+return OBJECT as a blessed reference of package Emacs::Lisp::Object.")
+     (object)
+     Lisp_Object object;
+{
+  return perlmacs_lisp_wrap_sv (perlmacs_lisp_to_sv (object));
+}
+
 static Lisp_Object
-internal_perl_call (nargs, args, do_eval)
+internal_perl_call (nargs, args, do_eval, conv)
      int nargs;
      Lisp_Object *args;
      int do_eval;
+     Lisp_Object (*conv) P_ ((SV *sv));
 {
   SV *sv;
   Lisp_Object code;
@@ -1448,28 +1564,23 @@ internal_perl_call (nargs, args, do_eval)
     PUTBACK;
   }
   if (do_eval)
-    /* FIXME: char * masquerading as SV *.  */
-    return call_perl (EVAL_AND_CALL, (SV *) XSTRING (code)->data, ctx);
+    return call_perl (EVAL_AND_CALL, XSTRING (code)->data, ctx);
   else
-    return call_perl (PERL_CALL_SV,
-		      sv_2mortal (perlmacs_lisp_to_sv (code)), ctx);
+    return call_perl (PERL_CALL_SV, sv_2mortal (perlmacs_lisp_to_sv (code)),
+		      ctx, conv);
 }
 
-DEFUN ("perl-eval", Fperl_eval, Sperl_eval, 1, 2, 0,
-  "Evaluate STRING as Perl code, returning the value of the last expression.\n\
-If specified, CONTEXT must be either `scalar-context', `list-context', or\n\
-`void-context'.  By default, a scalar context is supplied.")
-     (string, context)
+static Lisp_Object
+internal_perl_eval (string, context, conv)
      Lisp_Object string, context;
+     Lisp_Object (*conv) P_ ((SV *sv));
 {
   SV *sv;
   I32 ctx;
   struct perl_frame pf;
 
-  INIT_PERL;
-#if 0  /* Let's see if Perl can eval things other than strings.  :) */
   CHECK_STRING (string, 0);
-#endif
+  INIT_PERL;
 
   if (EQ (context, Qnil) || EQ (context, Qscalar_context))
     ctx = G_SCALAR;
@@ -1484,7 +1595,27 @@ If specified, CONTEXT must be either `scalar-context', `list-context', or\n\
 
   push_perl_frame (&pf, 1);
   sv = sv_2mortal (perlmacs_lisp_to_sv (string));
-  return call_perl (PERL_EVAL_SV, sv, ctx);
+  return call_perl (PERL_EVAL_SV, sv, ctx, conv);
+}
+
+DEFUN ("perl-eval", Fperl_eval, Sperl_eval, 1, 2, 0,
+  "Evaluate STRING as Perl code, returning the value of the last expression.\n\
+If specified, CONTEXT must be either `scalar-context', `list-context', or\n\
+`void-context'.  By default, a scalar context is supplied.")
+     (string, context)
+     Lisp_Object string, context;
+{
+  return internal_perl_eval (string, context, perlmacs_sv_to_lisp);
+}
+
+DEFUN ("perl-eval-raw", Fperl_eval_raw, Sperl_eval_raw, 1, 2, 0,
+  "Evaluate STRING as Perl code, returning its value as Perl data.\n\
+This function is exactly the same as `perl-eval' except in that it does not\n\
+convert its result to Lisp.")
+     (string, context)
+     Lisp_Object string, context;
+{
+  return internal_perl_eval (string, context, perlmacs_lisp_wrap_sv);
 }
 
 DEFUN ("perl-call", Fperl_call, Sperl_call, 1, MANY, 0,
@@ -1507,7 +1638,22 @@ the sub or coderef.\n\
      Lisp_Object *args;
 {
   INIT_PERL;
-  return internal_perl_call (nargs, args, 0);
+  return internal_perl_call (nargs, args, 0, perlmacs_sv_to_lisp);
+}
+
+DEFUN ("perl-call-raw", Fperl_call_raw, Sperl_call_raw, 1, MANY, 0,
+  "Call a Perl sub or coderef and return its result as Perl data.\n\
+This function is exactly the same as `perl-call' except in that it does not\n\
+convert its result to Lisp.\n\
+\n\
+\n\
+(perl-call-raw SUB &optional CONTEXT &rest ARGS)")
+  (nargs, args)
+     int nargs;
+     Lisp_Object *args;
+{
+  INIT_PERL;
+  return internal_perl_call (nargs, args, 0, perlmacs_lisp_wrap_sv);
 }
 
 DEFUN ("perl-eval-and-call", Fperl_eval_and_call, Sperl_eval_and_call,
@@ -1526,7 +1672,7 @@ sub name or coderef.  The remaining arguments are treated the same as in\n\
 {
   INIT_PERL;
   CHECK_STRING (args [0], 0);
-  return internal_perl_call (nargs, args, 1);
+  return internal_perl_call (nargs, args, 1, 0);
 }
 
 void
@@ -1546,8 +1692,6 @@ syms_of_perlmacs ()
   Qbad = intern ("bad");
   staticpro (&Qbad);
 
-  Qperl_call = intern ("perl-call");
-  staticpro (&Qperl_call);
   Qvoid_context = intern ("void-context");
   staticpro (&Qvoid_context);
   Qscalar_context = intern ("scalar-context");
@@ -1557,21 +1701,9 @@ syms_of_perlmacs ()
 
   Qperl_interpreter = intern ("perl-interpreter");
   staticpro (&Qperl_interpreter);
-  Qperl_object = intern ("perl-object");
-  staticpro (&Qperl_object);
 
-  Qperl_scalar = intern ("perl-scalar");
-  staticpro (&Qperl_scalar);
-  Qperl_array = intern ("perl-array");
-  staticpro (&Qperl_array);
-  Qperl_hash = intern ("perl-hash");
-  staticpro (&Qperl_hash);
-  Qperl_code = intern ("perl-code");
-  staticpro (&Qperl_code);
-  Qperl_glob = intern ("perl-glob");
-  staticpro (&Qperl_glob);
-
-  perl_interpreter = Qnil;
+  if (! top_level_perl)
+    perl_interpreter = Qnil;
   staticpro (&perl_interpreter);
 
   Qperl_error = intern ("perl-error");
@@ -1582,14 +1714,22 @@ syms_of_perlmacs ()
   Fput (Qperl_error, Qerror_message,
 	build_string ("Perl error"));
 
-  Fprovide (intern ("perl"));
+  Fprovide (intern ("perl-core"));
 
-  defsubr (&Sperl_eval);
-  defsubr (&Sperl_call);
-  defsubr (&Sperl_eval_and_call);
-  defsubr (&Sprimitive_make_perl);
   defsubr (&Sget_perl_interpreter);
   defsubr (&Sset_perl_interpreter);
+  defsubr (&Sprimitive_make_perl);
   defsubr (&Sperl_run);
   defsubr (&Sperl_destruct);
+  defsubr (&Sperl_status);
+  defsubr (&Sperl_phase);
+  defsubr (&Sperl_object_p);
+  defsubr (&Sperl_to_lisp);
+  defsubr (&Sperl_wrap);
+  defsubr (&Smake_perl_data);
+  defsubr (&Sperl_eval);
+  defsubr (&Sperl_eval_raw);
+  defsubr (&Sperl_call);
+  defsubr (&Sperl_call_raw);
+  defsubr (&Sperl_eval_and_call);
 }
